/**
*   Please feel free to use any fragment of the code in this file that you need
*   in your own work. As far as I am concerned, it's in the public domain. No
*   permission is necessary or required. Credit is always appreciated if you
*   use a large chunk or base a significant product on one of my examples,
*   but that's not required either.
*
*   This code is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
*
*      --- Joseph A. Huwaldt
**/

import gl4java.GLContext;
import java.awt.*;
import java.applet.*;
import java.awt.event.*;
import java.text.*;
import java.util.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.border.*;
import jahuwaldt.swing.*;


/**
*  A combination application/applet that displays the Pickover B
*  strange attractor and allows the user the interact with it.
*
*  <p>  Modified by:  Joseph A. Huwaldt   </p>
*
*  @author  Joseph A. Huwaldt   Date:  January 7, 2001
*  @version June 10, 2002
**/
public class PickoverApp extends JApplet {
  
  	//	The GL4Java canvas that actually displays the attractor.
	PickoverCanvas canvas = null;

	// The resolution of the slider scale.
	private static final int kSliderScale = (Integer.MAX_VALUE) / 100;

	// The tick mark increment for the slider scale.
	private static final int kTickSpacing = kSliderScale / 10;

	private static final int kNumParams = 5;
	
	// A copy of the sliders used in this window, one for each dimension.
	private JSlider[] sliders = new JSlider[kNumParams];

	// A copy of the text fields used by this window, one for each dimension.
	private DecimalField[] textFields = new DecimalField[kNumParams];

	// The minimum and maximum limits for the parameters.
	private float[] minBP = { 1, 0.1F, -3, -3.5F, -2 };
	private float[] maxBP = { 5, 2, 0, -1.5F, 2 };

	private String[] paramNames = {	"A", "B", "C", "D", "E" };
	
	public void init() {
		init(500,200);
	}

	/**
	*  Initialize this application/applet.
	**/
	public void init(int w, int h) {

		//	Set the system look and feel.
		try {
			setSystemLAF();
		} catch (Exception ex) {
			ex.printStackTrace();
		}
		
		Container cp = getContentPane();
		
		cp.setLayout(new BorderLayout());
		
		//	Create top panel.
		Box topPanel = new Box(BoxLayout.X_AXIS);
		cp.add("North", topPanel);
		
		//	Create the info button on the top of the window.
		JButton bInfo = new JButton("Info");
		bInfo.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent evt) {
					InfoWin win = new InfoWin();
				}
			} );
		
		topPanel.add(bInfo);
		topPanel.add(Box.createHorizontalGlue());
		
		//	Add the GLCanvas to the center of the window.
		canvas = new PickoverCanvas(w, h);
	
	    canvas.requestFocus();
		cp.add("Center", canvas);


		// Define the format we are going to use for the text fields.
		DecimalFormat format = (DecimalFormat) NumberFormat.getNumberInstance();
		format.setMaximumFractionDigits( 4 );
		format.setMinimumFractionDigits( 1 );
			
		// Create some event listeners.
		SliderListener sliderL = new SliderListener();
		TextFieldListener textL = new TextFieldListener();
		TextFieldFocusL focusL = new TextFieldFocusL();
		
		//	Create some parameter sliders.
		JPanel sliderPanel = new JPanel( new GridLayout( 1, 5 ) );
		sliderPanel.setBorder( BorderFactory.createEmptyBorder( 10, 10, 10, 10 ) );
		sliderPanel.setAlignmentX( JPanel.CENTER_ALIGNMENT );
		
		for (int i=0; i < kNumParams; ++i) {
			// Create a slider for this parameter variable.
			double value = getCurrentValue(i);
			int iValue = scaleToSlider(value, i);
			JSlider slider = new JSlider( JSlider.VERTICAL, 0, kSliderScale, iValue );
			slider.setMajorTickSpacing( kTickSpacing );
			slider.setPaintTicks( true );
			slider.addChangeListener( sliderL );
			Integer dimension = new Integer( i );
			slider.putClientProperty( "dimension", dimension );

			//Create labels for the top and bottom of the slider.
			Hashtable labelTable = new Hashtable();
			labelTable.put( new Integer( 0 ), new JLabel( Float.toString( minBP[i] ) ) );
			labelTable.put( new Integer( kSliderScale ),
									new JLabel( Float.toString( maxBP[i] ) ) );
			slider.setLabelTable( labelTable );
			slider.setPaintLabels( true );
			slider.setBorder( BorderFactory.createEmptyBorder( 0, 10, 5, 0 ) );
			sliders[i] = slider;
			
			// Create a text area for this slider.
			DecimalField textField = new DecimalField( value, 7, format );
			textField.addActionListener( textL );
			textField.addFocusListener( focusL );
			textField.putClientProperty( "dimension", dimension );
			textFields[i] = textField;
			
			// Layout the parameter panel.
			JPanel aPanel = new JPanel( new BorderLayout() );
			TitledBorder tborder = BorderFactory.createTitledBorder( paramNames[i] );
			tborder.setTitleJustification( TitledBorder.CENTER );
			aPanel.setBorder( tborder );
			aPanel.add( "Center", slider );
			aPanel.add( "South", textField );
				
			// Add the independent panel to the window's grid layout.
			sliderPanel.add( aPanel );
		}
		
		// Make it so you can tab from on text field to another.
		for ( int i = 0; i < kNumParams - 1; ++i )
			textFields[i].setNextFocusableComponent( textFields[i + 1] );
		textFields[kNumParams - 1].setNextFocusableComponent( textFields[0] );
		
		//	Add the sliders and text fields to the content pane.
		cp.add("South", sliderPanel);
		
		// Define the format we are going to use for the iteration # text field.
		format = (DecimalFormat) NumberFormat.getNumberInstance();
		format.setParseIntegerOnly(true);
			
		//	Create the number of iterations panel.
		JPanel iterPanel = new JPanel();
		topPanel.add(iterPanel);
		iterPanel.add(new JLabel("iterations = "));
		DecimalField textField = new DecimalField( canvas.iterations(), 7, format );
		textField.addActionListener( new ActionListener() {
				public void actionPerformed( ActionEvent e ) {
					DecimalField input = (DecimalField) e.getSource();
					int newIterations = (int)input.getValue();
					if (newIterations != canvas.iterations())
						canvas.setIterations(newIterations);
				}
			} );
		iterPanel.add(textField);

	} 

	public void destroy() {
	    canvas.destroy();
	}

	public void start() {
	}

	public void stop() {
	}

	/**
	*  The main method for this application if it is
	*  running as an application.
	**/
	public static void main( String args[] ) {
		String gljLib=null;
		String glLib=null;
		String gluLib=null;

	    if(args.length > 0)
			gljLib = args[0];
	    if(args.length > 1)
			glLib = args[1];
	    if(args.length > 2)
			gluLib = args[2];

		if(GLContext.loadNativeLibraries(gljLib, glLib, gluLib)==false) {
			System.out.println("Could not load native GL4Java libs:"+
									gljLib + ", " + glLib + ", " + gluLib);
			return;
		}

		GLContext.gljNativeDebug = false;
		GLContext.gljClassDebug  = false;

		PickoverApp window = new PickoverApp();

		JFrame f = new JFrame();
		f.getContentPane().add(window);
		window.init();
		f.pack();
		f.setVisible(true);

		f.addWindowListener( new WindowAdapter() {
				public void windowClosed(WindowEvent e) {
					System.exit(0);
				}
				public void windowClosing(WindowEvent e) {
					System.exit(0);
				}
				}
			);


		window.start();
	}

	/**
	*  Set the system look and feel to hide that hideous Java LAF.
	**/
	private void setSystemLAF() throws Exception {
		// Set the system look and feel to hide that hideous Java LAF.
		String laf = UIManager.getSystemLookAndFeelClassName();
		UIManager.setLookAndFeel( laf );
	}
	
	/**
	*  Returns the value of the specified parameter.
	**/
	private double getCurrentValue(int paramIdx) {
		double value = 0;
		switch(paramIdx) {
			case 0:	value = canvas.getA();
					break;
			case 1:	value = canvas.getB();
					break;
			case 2:	value = canvas.getC();
					break;
			case 3:	value = canvas.getD();
					break;
			case 4:	value = canvas.getE();
					break;
		}
		
		return value;
	}
	
	private int scaleToSlider(double value, int paramIdx) {
		int val = (int) ((value - minBP[paramIdx])/(maxBP[paramIdx] - minBP[paramIdx])*kSliderScale);
		return val;
	}
	
	/**
	*  Defines a listener for our sliders that is called whenever the slider
	*  is changed.
	**/
	class SliderListener extends Object implements ChangeListener {

		public void stateChanged( ChangeEvent evt ) {
			JSlider input = (JSlider) evt.getSource();
			
			// Find out what dimension this is and what value it has been set to.
			int i = ((Integer) input.getClientProperty( "dimension" )).intValue();
			int val = input.getValue();
			
			// Update the text field to match the slider.
			double value = (maxBP[i] - minBP[i])*val/kSliderScale + minBP[i];
			textFields[i].setValue( value );
			
			double a = canvas.getA();
			double b = canvas.getB();
			double c = canvas.getC();
			double d = canvas.getD();
			double e = canvas.getE();
			switch(i) {
				case 0:	a = value;
						break;
				case 1:	b = value;
						break;
				case 2:	c = value;
						break;
				case 3:	d = value;
						break;
				case 4:	e = value;
						break;
			}
			
			// Change the Pickover attractor.
			canvas.setInputs(a, b, c, d, e);
		}
	}
	
	/**
	*  Handles the user changing the value in a text field.
	*
	*  @param  input  The text field that has had it's value changed.
	**/
	private void doTextFieldChange( DecimalField input ) {
		
		// Find out what dimension this is and what value it has been set to.
		int i = ((Integer) input.getClientProperty( "dimension" )).intValue();
		double value = (double) input.getValue();
		
		// Validate the user's input.
		if ( value > maxBP[i] )
			value = maxBP[i];
		
		else {
			if ( value < minBP[i] )
				value = minBP[i];
		}
		
		// Update the slider to match the text input.
		int val = (int) ((value - minBP[i])/(maxBP[i] - minBP[i])*kSliderScale);
		sliders[i].setValue( val );
		
		double a = canvas.getA();
		double b = canvas.getB();
		double c = canvas.getC();
		double d = canvas.getD();
		double e = canvas.getE();
		switch(i) {
			case 0:	a = value;
					break;
			case 1:	b = value;
					break;
			case 2:	c = value;
					break;
			case 3:	d = value;
					break;
			case 4:	e = value;
					break;
		}
		
		// Change the Pickover attractor.
		canvas.setInputs(a, b, c, d, e);
	}

	/**
	*  Defines a listener for our text fields that is called whenever the user
	*  types in a new number and presses enter.
	**/
	class TextFieldListener extends Object implements ActionListener {

		public void actionPerformed( ActionEvent e ) {
			DecimalField input = (DecimalField) e.getSource();
			doTextFieldChange( input );
		}


	}
	/**
	*  Defines a listener for our text fields that is called whenever the
	*  focus changes off a particular text field.
	**/
	class TextFieldFocusL extends Object implements FocusListener {

		public void focusGained( FocusEvent e ) { }

		public void focusLost( FocusEvent e ) {
			DecimalField input = (DecimalField) e.getSource();
			doTextFieldChange( input );
		}
	}

	/**
	*  This is an information window that shows information about
	*  this application.
	**/
	class InfoWin extends JFrame implements ActionListener {
		protected JButton DismissButton;

		String text =
			"  Pickover, Version 1.0.1\n"+
			"     by Joseph A. Huwaldt\n\n"+
			"\thttp://homepage.mac.com/jhuwaldt/java/index.html\n"+
			"\tmailto:jhuwaldt@knology.net\n\n"+
			"  Use the mouse to rotate the 3D view of the Pickover B attractor.\n"+
			"  Attractor is formed by iterating the following equations:\n\n"+
			"\tx(n+1) = sin(a*y(n)) - z(n)*cos(b*x(n))\n"+
			"\ty(n+1) = z(n)*sin(c*x(n) - cos(d*y(n))\n"+
			"\tz(n+1) = e*sin(x(n))\n";

		public InfoWin() {
			super("INFO");
			Container cp = getContentPane();
			
			// Define a window listener that handles closing this window.
			addWindowListener( new WindowAdapter() {
					public void windowClosing( WindowEvent e ) {
						// Dispose of this window.
						dispose();
					}
				} );

			cp.setLayout(new BorderLayout());
			cp.add("Center", new JTextArea(text, 25, 50));

			cp.add("South", DismissButton = new JButton("Close"));
			DismissButton.addActionListener(this);
			
			pack();
			show();
		}

		public void actionPerformed(ActionEvent event) {
			if ( event.getSource() == DismissButton) {
			    dispose();
			}
		}
	}
	
}
