/**
*   Please feel free to use any fragment of the code in this file that you need
*   in your own work. As far as I am concerned, it's in the public domain. No
*   permission is necessary or required. Credit is always appreciated if you
*   use a large chunk or base a significant product on one of my examples,
*   but that's not required either.
*
*   This code is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
*
*      --- Joseph A. Huwaldt
**/

import gl4java.GLContext;
import gl4java.awt.GLCanvas;
import java.awt.*;
import java.awt.event.*;
import java.lang.*;
import java.util.*;
import jahuwaldt.gl.Matrix;


/**
*  An OpenGL Canvas that displays the Pickover B
*  strange attractor and allows the user to rotate it
*  around with the mouse.
*
*  <p>  Modified by:  Joseph A. Huwaldt   </p>
*
*  @author  Joseph A. Huwaldt   Date:  January 7, 2001
*  @version January 13, 2001
**/
public class PickoverCanvas extends GLCanvas {

	private int numIterations = 20000;
	
	//	Pickover attractor inputs.
	private double a = 2.24;
	private double b = 0.43;
	private double c = -0.65;
	private double d = -2.43;
	private double e = 1;
	
	private float pointList[][] = new float[numIterations][3];
	
	private int colorList[] = new int[numIterations];

	private final float colorSet[][] = {
			/* Shades of red. */
			{ 0.7f, 0.2f, 0.4f, 1f },
			{ 0.8f, 0.0f, 0.7f, 1f },
			{ 1.0f, 0.0f, 0.0f, 1f },
			{ 0.9f, 0.3f, 0.6f, 1f },
			{ 1.0f, 0.4f, 0.0f, 1f },
			{ 1.0f, 0.0f, 0.5f, 1f }
		};
	
	private final int kNumColors = colorSet.length;
	
	private int prevMouseX, prevMouseY;
	private boolean mouseRButtonDown = false;
	private float rot_matrix[] = Matrix.identity();

	
	/**
	*  Construct a GLCanvas that displays the Pickover B
	*  strange attractor.
	**/
	public PickoverCanvas(int w, int h) {
		super(w,h);
		GLContext.gljNativeDebug = false;
		GLContext.gljClassDebug = false;

		//	Create the pickover attractor points.
		createPoints();
		assignColors();
		
		addMouseListener(new MouseL());
		addMouseMotionListener(new MotionL());
	}

	/**
	*  Set the inputs to the pickover B strange attractor.
	**/
	public void setInputs(double A, double B, double C, double D, double E) {
		a = A;
		b = B;
		c = C;
		d = D;
		e = E;

		//	Create the pickover attractor points.
		createPoints();
		repaint();
	}
	
	/**
	*  Return the Pickover B Strange Attractor parameter "A".
	**/
	public double getA() {
		return a;
	}
	
	/**
	*  Return the Pickover B Strange Attractor parameter "B".
	**/
	public double getB() {
		return b;
	}
	
	/**
	*  Return the Pickover B Strange Attractor parameter "C".
	**/
	public double getC() {
		return c;
	}
	
	/**
	*  Return the Pickover B Strange Attractor parameter "D".
	**/
	public double getD() {
		return d;
	}
	
	/**
	*  Return the Pickover B Strange Attractor parameter "E".
	**/
	public double getE() {
		return e;
	}
	
	/**
	*  Return the number of iterations being used.
	**/
	public int iterations() {
		return numIterations;
	}
	
	/**
	*  Set the number of iterations to use.
	**/
	public void setIterations(int number) {
		numIterations = number;
		pointList = new float[numIterations][3];
		colorList = new int[numIterations];
		
		//	Create the pickover attractor points.
		createPoints();
		assignColors();
		repaint();
	}
	
	/**
	*  Create points for Pickover B strange attractor.
	**/
	private void createPoints() {
		double x = 0, y = 0, z = 0;
		
		for (int i=0; i < numIterations; ++i) {
			double xtemp = Math.sin(a*y) - z*Math.cos(b*x);
			y = z*Math.sin(c*x) - Math.cos(d*y);
			z = e*Math.sin(x);
			x = xtemp;
			
			pointList[i][0] = (float)x;
			pointList[i][1] = (float)y;
			pointList[i][2] = (float)z;
		}
	}
	
	/**
	*  Assign colors to each point in the attractor.
	**/
	private void assignColors() {
		for (int i=0; i < numIterations; ++i) {
			colorList[i] = (int)(Math.random()*(double)kNumColors);
		}
	}
	
	/**
	* void init()
	*
	* Called just AFTER the GL-Context is created.
	**/
	public void init() {
		super.init();
		
		//	Set initial rotation angles.
		setInitialRotation();
		
		//This Will Clear The Background Color To Black
		gl.glClearColor(0.0f, 0.0f, 0.0f, 0.0f);

		//Enables Clearing Of The Depth Buffer
		gl.glClearDepth(1.0);
		//The Type Of Depth Test To Do
		gl.glDepthFunc(GL_LESS);
		//Enables Depth Testing
		gl.glEnable(GL_DEPTH_TEST);

		//Enables Smooth Color Shading
		gl.glShadeModel(GL_SMOOTH);
		
		//	Set up fog.
		float black[] = { 0, 0, 0, 0 };
		gl.glEnable(GL_FOG);
		gl.glFogi(GL_FOG_MODE, GL_LINEAR);
		gl.glFogf(GL_FOG_DENSITY,0.5f);
		gl.glFogf(GL_FOG_START, 1.5f);
		gl.glFogf(GL_FOG_END, 7.5f);
		gl.glFogfv(GL_FOG_COLOR, black);
		gl.glHint (GL_FOG_HINT, GL_NICEST);  /*  per pixel   */

		gl.glPointSize(1);
//		gl.glEnable(GL_POINT_SMOOTH);
		gl.glEnable(GL_BLEND);
		gl.glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		gl.glEnable(GL_COLOR_MATERIAL);
		gl.glEnable(GL_LIGHT0);

		//	Initialize projection matrix, etc.
		Dimension dim = getSize();
		reshape(dim.width, dim.height);
		
	}

	/**
	* void destroy()
	*
	* Destroy the canvas.
	**/
	public void destroy() {
		//Destroy the GLContext
		cvsDispose();
	}
	
	/**
	* void reshape(int width, int height)
	*
	* Called after the first paint command.
	**/
	public void reshape(int width, int height) {
		//Reset The Current Viewport And Perspective Transformation
		gl.glViewport(0, 0, width, height);

		//Select The Projection Matrix
		gl.glMatrixMode(GL_PROJECTION);
		//Reset The Projection Matrix
		gl.glLoadIdentity();
		
		//Calculate The Aspect Ratio Of The Window
		glu.gluPerspective(45.0f, (float)getSize().width / (float)getSize().height, 0.1f, 10.0f);
		
		//Select The Modelview Matrix
		gl.glMatrixMode(GL_MODELVIEW);
		gl.glLoadIdentity();
		
		glu.gluLookAt(	0,0,+5,
						0,0,1,
						0.0,1.0,0.0);

	}

	/**
	* void display()
	*
	* Draw to the canvas.
	**/
	public void display() {
		//	Ensure GL is initialised correctly
		if (!cvsIsInit() || glj.gljMakeCurrent() == false)	return;
		
		//	Clear The Screen And The Depth Buffer
		gl.glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

		//	Set The View
		gl.glPushMatrix();
		gl.glMultMatrixf(rot_matrix);

		//	Allow points to blend with each other.
		gl.glDepthMask(true);
		
		//	Draw all the points.
		gl.glBegin(GL_POINTS);
		for (int i = 0; i < numIterations; i++) {
			int c = colorList[i];
			gl.glColor4fv(colorSet[colorList[i]]);
            gl.glVertex3fv(pointList[i]);
		}
		gl.glEnd();
		
		gl.glPopMatrix();
		
		//	Do the animation.
		glj.gljSwap();
		glj.gljCheckGL();
		glj.gljFree();
	}

	
	/**
	*  Set initial rotation angles.
	**/
	private void setInitialRotation() {

    	Matrix.rotateY(Matrix.deg2Rad(30), rot_matrix);
    	Matrix.rotateX(Matrix.deg2Rad(20), rot_matrix);
    	
	}

	/**
	*  A mouse listener that looks for the right mouse button being used.
	**/
	class MouseL implements MouseListener {
		
		public void mouseEntered( MouseEvent evt ) { }
    
		public void mouseExited( MouseEvent evt ) { }
    
		public void mousePressed( MouseEvent evt ) {
			prevMouseX = evt.getX();
			prevMouseY = evt.getY();
			if ((evt.getModifiers() & evt.BUTTON3_MASK) != 0) {
				mouseRButtonDown = true;
				evt.consume();
			}
		}

		public void mouseReleased( MouseEvent evt ) {
			if ((evt.getModifiers() & evt.BUTTON3_MASK) != 0) {
				mouseRButtonDown = false;
				evt.consume();
			}
		}

		public void mouseClicked( MouseEvent evt ) { }
	}
	
	/**
	*  A mouse motion listener that allows the user to rotate the model.
	**/
	class MotionL implements MouseMotionListener {
		
		public void mouseDragged( MouseEvent evt ) {
			int x = evt.getX();
			int y = evt.getY();
			float thetaX = (float)(x - prevMouseX)*(360.0f/(float)getSize().width);
			float thetaY = (float)(prevMouseY - y)*(360.0f/(float)getSize().height);

			prevMouseX = x;
			prevMouseY = y;
			if ( (thetaX != 0.0f) || (thetaY != 0.0f) ) {
				if (mouseRButtonDown)
					Matrix.rotateZ(Matrix.deg2Rad(thetaY), rot_matrix);
				else
					Matrix.rotateX(Matrix.deg2Rad(-thetaY), rot_matrix);

				Matrix.rotateY(Matrix.deg2Rad(thetaX), rot_matrix);
			}
			evt.consume();
			
			repaint();
		}

		public void mouseMoved( MouseEvent evt ) { }
	}
	
}

